import inspect

from globus_compute_sdk import Executor, Client
from custom_image_builder.exception.ImageBuilderException import ImageBuilderException
from custom_image_builder.exception.RegisterImageException import RegisterImageException


def funcx_build_image(image_file_name: str,
                      base_image_type: str,
                      base_image: str,
                      payload_url: str,
                      pip_packages: list,
                      conda_packages: list,
                      apt_packages: list,
                      images_dir: str) -> (str, str):
    """Builds apptainer/singularity image on HPC cluster using funcX, return path of the image and logs

    :param apt_packages:
    :param conda_packages:
    :param pip_packages:
    :param payload_url:
    :param base_image:
    :param image_file_name:
    :param base_image_type:
    :param images_dir:
    :return: logs, path_of_image
    """
    from jinja2 import Template
    import subprocess
    import os

    build_file = """BootStrap: {{ container_type }}
From: {{ base_image }}

%post
    {% if apt is not none and apt|length > 0 %}
        {% for package in apt %}
        apt-get install {{ package }}
        {% endfor %}
    {% endif %}

    {% if payload_url is not none %}
        git clone {{payload_url }}
    {% endif %}

    {% if pip is not none and pip|length > 0 %}
        {% for package in pip %}
        pip install {{ package }}
        {% endfor %}
    {% endif %}

    {% if conda is not none and conda|length > 0 %}
        {% for package in conda %}
        conda install {{ package }}
        {% endfor %}
    {% endif %}
    
    pip install globus-compute-endpoint==2.2.0
    pip install globus-compute-common==0.2.0

%runscript
    """

    template_str = Template(build_file)

    build_file = template_str.render({
        "container_type": base_image_type,
        "base_image": base_image,
        "payload_url": payload_url,
        "pip": pip_packages,
        "apt": apt_packages,
        "conda": conda_packages
    })

    build_image_file_name = image_file_name
    open(f"{build_image_file_name}.def", "w").write(build_file)

    # create directory if not exists
    os.makedirs(images_dir, exist_ok=True)

    process = subprocess.Popen(
        f"apptainer build --force {images_dir}/{build_image_file_name}.sif {build_image_file_name}.def",
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        shell=True
    )

    exit_code = process.wait()
    stdout, stderr = process.communicate()

    return f"Output logs are as follows: \n {str(stderr)}  \n {str(stdout)}", f"{images_dir}/{build_image_file_name}.sif", exit_code


def register_container(image_file_path, gcc):
    try:
        # Register sif file path, created after successfully building the image
        container_id = gcc.register_container(image_file_path, "singularity")

        return container_id

    except Exception as ex:
        raise RegisterImageException(ex)

def redef_in_main(obj):
    """Helper: redefine an object in __main__

    This has the effect of coaxing dill into serializing both the definition and
    the instance of an object together (in the case of a class), so it can be
    deserialized without needing the definition to be available for import on
    the other side. We do this for the "real" function we register, too (see below)
    in order to guarantee that there are no intra-garden references that dill might
    try to import on the other end.

    This works because dill is smart enough to know that if you defined a class
    interactively (like in a repl) then it can't expect that definition to be
    available after the session exits, so has to include it.
    """

    # make sure it's not already in main
    if obj.__module__ != "__main__":
        import __main__

        s = inspect.getsource(obj)
        exec(s, __main__.__dict__)

def build_image(gc_executor: Executor,
                image_file_name: str,
                base_image_type: str,
                base_image: str,
                payload_url: str,
                pip_packages: list,
                conda_packages: list,
                apt_packages: list,
                images_dir: str) -> str:
    """Calls Globus compute executor which Builds apptainer/singularity image on HPC cluster using funcX, return path of the image and logs

    :param gc_executor:
    :param apt_packages:
    :param conda_packages:
    :param pip_packages:
    :param payload_url:
    :param base_image:
    :param image_file_name:
    :param base_image_type:
    :param images_dir:
    :return: image_file_path
    """
    import __main__

    try:
        with gc_executor as ex:
            redef_in_main(funcx_build_image)
            fut = ex.submit(__main__.funcx_build_image,
                                     image_file_name,
                                     base_image_type,
                                     base_image,
                                     payload_url,
                                     pip_packages,
                                     conda_packages,
                                     apt_packages,
                                     images_dir)

        logs, image_file_path, exit_code = fut.result()

        if exit_code != 0:
            raise ImageBuilderException(f"Failed to build image,  {logs}")

        return image_file_path
    except Exception as ex:
        raise ImageBuilderException(ex)


def build_and_register_container(gcc_client: Client,
                                 endpoint_id: str,
                                 image_file_name: str,
                                 base_image_type: str,
                                 base_image: str,
                                 images_dir: str,
                                 payload_url: str = None,
                                 pip_packages: list = None,
                                 conda_packages: list = None,
                                 apt_packages: list = None) -> str:
    """Calls Globus compute executor and client which Builds apptainer/singularity image on HPC cluster using funcX, and
    registers the image returning the container id

        :param gcc_client:
        :param endpoint_id:
        :param apt_packages:
        :param conda_packages:
        :param pip_packages:
        :param payload_url:
        :param base_image:
        :param images_dir:
        :param image_file_name:
        :param base_image_type:
        :return: container_id """
    gc_executor = Executor(endpoint_id=endpoint_id, funcx_client=gcc_client)

    image_file_path = build_image(gc_executor,
                                  image_file_name,
                                  base_image_type,
                                  base_image,
                                  payload_url,
                                  pip_packages,
                                  conda_packages,
                                  apt_packages,
                                  images_dir)

    return register_container(image_file_path, gcc_client)



import click
import json

from diamond.container.container import ImageRegistry

SITE_CONFIG_PATH = "/tmp/site_config.json"


def init_ir() -> ImageRegistry:
    with open(SITE_CONFIG_PATH, 'r') as f:
        site_config = json.load(f)
    ir = ImageRegistry(site_config)
    return ir

@click.group()
def cli():
    pass

@cli.command()
@click.argument('endpoint_id')
@click.argument('work_path')
def setup_endpoint(endpoint_id: str, work_path: str):
    site_config = {
        "endpoint_id": endpoint_id,
        "work_path": work_path
    }
    with open(SITE_CONFIG_PATH, 'w') as f:
        json.dump(site_config, f)

@cli.command()
def list_base_image_tags():
    ir = init_ir()
    print(ir.list_base_image_tags())

@cli.command()
def list_avail_containers():
    ir = init_ir()
    print(ir.list_avail_containers())

@cli.command()
@click.argument('img_file_name')
def exists(img_file_name: str):
    ir = init_ir()
    print(ir.exists(img_file_name))

@cli.command()
@click.argument('base_img')
@click.argument('img_file_name')
@click.argument('pkg_config_path')
@click.option('--force', is_flag=False)
def build(base_img: str, pkg_config_path: str, img_file_name: str, force: bool):
    ir = init_ir()
    container_id = ir.build(base_img, pkg_config_path, img_file_name, force)
    if container_id:
        print(f"The container id is {container_id}")
    else:
        print("Failed to build the container.")

if __name__ == '__main__':
    cli()



from globus_compute_sdk import Client, Executor
from .custom_image_builder import build_and_register_container
from .custom_image_builder.image_builder import redef_in_main
import traceback
import os
import requests
import json
from typing import Union

class ImageRegistry:
    """
    This class provides methods for building images, checking if an image exists in the registry, and listing containers.
    """

    IMAGES_FOLDER = ".images"
    REGISTRY_FILE_NAME = "registry.json"
    BASE_IMAGE_CATALOG_URL = "https://catalog.ngc.nvidia.com/api/containers/images?orgName=nvidia&name=pytorch&isPublic=true"
    BASE_IMAGE_PREFIX = "nvcr.io/nvidia/pytorch:"

    def __init__(self, site_config: dict):
        """
        Initialize the ImageRegistry with a specific site configuration.

        :param site_config: The config for an endpoint
        """
        self.site_config = site_config
        self.gcc_client = Client()
        self.endpoint = site_config['endpoint']
        self.images_dir = os.path.join(site_config['work_path'], self.IMAGES_FOLDER)
        self.registry_file_path = os.path.join(self.images_dir, self.REGISTRY_FILE_NAME)
        print("Loading the image registry...")
        # the registry is a dictionary with image file names as keys and container ids as values
        import __main__
        with Executor(endpoint_id=self.endpoint) as ex:
            redef_in_main(funcx_get_registry)
            fut = ex.submit(__main__.funcx_get_registry, self.registry_file_path)
        self.registry = fut.result()

    def build(self, base_img: str, pkg_config: Union[str, list], img_file_name: str, base_img_type: str = "docker", force: bool = False) -> str:
        """
        This function lets users build an apptainer image with a base image and a required package config on an endpoint.

        :param base_img: The name of the base image
        :param pkg_config: Pip package configuration, a file path or a list of packages
        :param img_file_name: The name of the image file
        :param base_img_type: The type of the base image
        :param force: If True, the image will be built even if it already exists
        :return: The built container id
        """
        if not force and self.exists(img_file_name):
            print(f"The image [{img_file_name}] already exists in the registry.")
            return self.registry[img_file_name]
        
        pip_packages = []
        if isinstance(pkg_config, list):
            pip_packages = pkg_config
        elif isinstance(pkg_config, str):
            if not os.path.exists(pkg_config):
                print("The package config file does not exist.")
            else:
                with open(pkg_config, 'r') as f:
                    pip_packages = f.read().strip().split("\n")
        print(f"The pip packages to be installed are {pip_packages}")

        try:
            container_id = build_and_register_container(gcc_client=self.gcc_client,
                                                        endpoint_id=self.endpoint,
                                                        image_file_name=img_file_name,
                                                        base_image_type=base_img_type,
                                                        base_image=base_img,
                                                        images_dir=self.images_dir,
                                                        pip_packages=pip_packages
                                                        )
            # update the image registry
            self.update_registry(img_file_name, container_id)
            return container_id
        except Exception as e:
            traceback.print_exc()
            return None

    def exists(self, img_file_name: str) -> bool:
        """
        This function checks if an image is already in the local registry.

        :param img_file_name: The name of the image file
        :return: True if the image is in the registry, otherwise False
        """
        return img_file_name in self.registry
    
    def update_registry(self, img_file_name: str, container_id: str):
        """
        This function updates the local registry with a new image.

        :param img_file_name: The name of the image file
        :param container_id: The container id
        """
        self.registry[img_file_name] = container_id
        import __main__
        with Executor(endpoint_id=self.endpoint) as ex:
            redef_in_main(funcx_update_registry)
            fut = ex.submit(__main__.funcx_update_registry, self.registry_file_path, self.registry)
        # wait for the update to finish
        fut.result()

    def list_avail_containers(self):
        """
        This function lists all the available Apptainer containers on an endpoint.
        """
        return self.registry

    def list_base_image_tags(self):
        """
        This function lists all the available base image tags from a third-party registry.
        """
        images_info = json.loads(requests.get(self.BASE_IMAGE_CATALOG_URL).text)['images']
        tags = [x['tag'] for x in images_info]
        full_tags = [f"{self.BASE_IMAGE_PREFIX}{x}" for x in tags]
        return full_tags
    

def funcx_get_registry(registry_file_path):
    """
    This function is supposed to be submitted by the client and returns the registry dictionary.

    :param registry_file_path: The path to the registry file on the endpoint
    :return: The registry dictionary
    """
    import json
    import os

    registry = {}
    if os.path.exists(registry_file_path):
        # load the registry json file
        with open(registry_file_path, 'r') as f:
            registry = json.load(f)
    return registry


def funcx_update_registry(registry_file_path, registry):
    """
    This function is supposed to be submitted by the client and updates the registry file on the endpoint.

    :param registry_file_path: The path to the registry file on the endpoint
    :param registry: The registry dictionary
    """
    import json

    with open(registry_file_path, 'w') as f:
        json.dump(registry, f)


if __name__ == "__main__":
    site_config = {
        "endpoint": "847f59fb-7286-452c-b490-0270ddd8da94",
        "work_path": "/work2/09849/gcyang/frontera"
    }
    ir = ImageRegistry(site_config)

    print("The available base image tags are:")
    print(ir.list_base_image_tags())

    print("The available containers are:")
    print(ir.list_avail_containers())

    # pick a base image from https://catalog.ngc.nvidia.com/orgs/nvidia/containers/pytorch/tags
    base_img = "nvcr.io/nvidia/pytorch:24.02-py3"
    # save an example of pip config to a file
    example_pkg_str = "lightning\nscipy\n"
    pkg_config_path = "/tmp/pip_config.txt"
    with open(pkg_config_path, 'w') as f:
        f.write(example_pkg_str)
    img_file_name = "nv_pytorch"
    container_id = ir.build(base_img, pkg_config_path, img_file_name)
    if container_id:
        print(f"The container id is {container_id}")
    else:
        print("Failed to build the container.")
